#ifndef __util_h
#define __util_h

#include <iostream>
#include <cstdlib>

// The functions in this file were based on the following paper:

// Gumerov, N., Duraiswami, R., and Borovikov, Y. (2003).  Data structures,
// optimal choice of parameters, and complexity results for generalized
// multilevel fast multipole methods in d dimensions.  Technical Report
// UMIACS-TR-2003-28, University of Maryland, College Park, Maryland.


inline int Parent(const int n)
{
   return n >> 3;
}

inline int GrandParent(int n, int levels)
{
   while(levels--)
      n >>= 3;

   return n;
}

inline void ChildrenAll(const int n, 
			int children[8])
{

   const int N = n << 3;
   for(int i=0;i<8;++i)
   {
      children[i] = N + i;
   }

}

// get the first child of the current box
inline int ChildFirst(const int n)
{

   return n << 3;

}



/* Recall that a 32 bit floating point number is represented as

S EEEEEEEE MMMMMMMMMMMMMMMMMMMMMMM

S = 1 Sign bit
E = 8 bit exponent
M = 23 bit mantissa (and the one before is implied)

1+8+23 = 32

where the number is equal to:

(-1)^S * 2^(E-127) * (1.M)

*/


/* Recall that a 64 bit floating point number is represented as

S EEEEEEEEEEE MMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMM

S = 1 Sign bit
E = 11 bit exponent
M = 52 bit mantissa (and the one before is implied)

1+11+52 = 64

where the number is equal to:

(-1)^S * 2^(E-1023) * (1.M)

*/


// Get the unbiased exponent part of the floating point value as an int. We
// assume the value is positive so we don't mask out the sign bit
inline int exponent(float f)
{

   const int n = *reinterpret_cast<int*>(&f);


   return (n >> 23) - 127;
   
}


// Get the unbiased exponent part of the double floating point value as an
// int. We assume the value is positive so we don't mask out the sign bit
inline int exponent(double f)
{

   const long long int n = *reinterpret_cast<long long int*>(&f);


   return static_cast<int>( (n >> 52) - 1023 );
   
}


inline int mantissa(float f)
{

   const int n = *reinterpret_cast<int*>(&f);
   

   // Mask out the sign and exponent bits
   // (Note: 0x007FFFFF = 11111111 11111111 01111111 00000000
   // in little endian)

   return n & 0x007FFFFF;
}

inline long long int mantissa(double f)
{

   const long long int n = *reinterpret_cast<long long int*>(&f);
   

   // Mask out the sign and exponent bits
   // (Note: 0x000FFFFFFFFFFFFFLL = 
   //   11111111 11111111 11111111 11111111 11111111 11111111 00001111 00000000
   // in little endian)

   return n & 0x000FFFFFFFFFFFFFLL;
}



inline int Interleave(const int x,
		      const int y,
		      const int z,
		      const int level)
{

#ifndef NDEBUG
   if(level>10)
   {
      std::cout << "level must be <= 10" << std::endl;
      abort();
   }
#endif


   int w = 0;


   // adapted from "Bit Twiddling Hacks" By Sean Eron Anderson
   // http://graphics.stanford.edu/~seander/bithacks.html
   for(int i=0;i<level;i++)
   {

      const int i2 = 2*i;

      w |= 
	 ( (z & (1<<i)) <<  i2      ) | 
	 ( (y & (1<<i)) << (i2 + 1) ) | 
	 ( (x & (1<<i)) << (i2 + 2) ) ;


   }

   return w;

}

inline long long int Interleave(const long long int x,
				const long long int y,
				const long long int z,
				const int level)
{

#ifndef NDEBUG
   if(level>10)
   {
      std::cout << "level must be <= 10" << std::endl;
      abort();
   }
#endif


   long long int w = 0LL;

   // adapted from "Bit Twiddling Hacks" By Sean Eron Anderson
   // http://graphics.stanford.edu/~seander/bithacks.html
   for(long long int i=0LL;i<static_cast<long long int>(level);i++)
   {

      const long long int i2 = 2LL*i;

      w |= 
	 ( (z & (1LL<<i)) <<  i2        ) | 
	 ( (y & (1LL<<i)) << (i2 + 1LL) ) | 
	 ( (x & (1LL<<i)) << (i2 + 2LL) ) ;


   }

   return w;

}


inline void Deinterleave(const int index,
			 const int level,
			 int* x,
			 int* y,
			 int* z)
				 
{

#ifndef NDEBUG
   if(level>10)
   {
      std::cout << "level must be <= 10" << std::endl;
      abort();
   }
#endif


   *x = *y = *z = 0;

   const int level3 = 3*level;
   int j=0;

   for(int i=0; i<level3; i+=3)
   {
      const int i_j = i - j;

      *z |= (index & (1 <<  i     )) >> (i_j    );
      *y |= (index & (1 << (i + 1))) >> (i_j + 1);
      *x |= (index & (1 << (i + 2))) >> (i_j + 2);

      ++j;
   }

}



// Get the box a given point is located in for a specific level.
// Note the point must be in the unit cube.
inline int Box(const float x,
	       const float y,
	       const float z,
	       const int level)
{



   // Place the implied 1 at the beginning of the mantissa
   // (i.e. bit 23)

   int xM = mantissa(x) | (1<<23);
   int yM = mantissa(y) | (1<<23);
   int zM = mantissa(z) | (1<<23);


   // Right shift the mantissa so that only the first level bits after the
   // "decimal" point remain

   const int _23_minus_level = 23 - level;


   // any shift amount greater than 23 will result in 0
   int shiftamount = (-exponent(x) + _23_minus_level);
   if(shiftamount > 23)
      xM = 0;
   else
      xM >>= shiftamount;

   shiftamount = (-exponent(y) + _23_minus_level);
   if(shiftamount > 23)
      yM = 0;
   else
      yM >>= shiftamount;

   shiftamount = (-exponent(z) + _23_minus_level);
   if(shiftamount > 23)
      zM = 0;
   else
      zM >>= shiftamount;

   
   return Interleave(xM,yM,zM,level);

}

// Get the box a given point is located in for a specific level.
// Note the point must be in the unit cube.
inline int Box(const double x,
	       const double y,
	       const double z,
	       const int level)
{



   // Place the implied 1 at the beginning of the mantissa
   // (i.e. bit 52)

   long long int xM = mantissa(x) | (1LL<<52LL);
   long long int yM = mantissa(y) | (1LL<<52LL);
   long long int zM = mantissa(z) | (1LL<<52LL);


   // Right shift the mantissa so that only the first level bits after the
   // "decimal" point remain

   const int _52_minus_level = 52 - level;


   // any shift amount greater than 52 will result in 0
   int shiftamount = (-exponent(x) + _52_minus_level);
   if(shiftamount > 52)
      xM = 0LL;
   else
      xM >>= shiftamount;

   shiftamount = (-exponent(y) + _52_minus_level);
   if(shiftamount > 52)
      yM = 0LL;
   else
      yM >>= shiftamount;

   shiftamount = (-exponent(z) + _52_minus_level);
   if(shiftamount > 52)
      zM = 0LL;
   else
      zM >>= shiftamount;

   
   return static_cast<int>( Interleave(xM,yM,zM,level) );

}


inline void Center(const int boxIndex, const int level, float center[3])
{
   Deinterleave(boxIndex,level,
		reinterpret_cast<int*>(&center[0]),
		reinterpret_cast<int*>(&center[1]),
		reinterpret_cast<int*>(&center[2]));

   const int _23_minus_level = 23 - level;

   for(int i=0;i<3;++i)
   {

      //int& x = *((int*)(&center[i]));
      int& x = *reinterpret_cast<int*>(&center[i]);

      // Shift 1 bit to left and place a 1 in the last bit (i.e. bit #1)
      x = (x << 1) | 1;
   
      // left shift
      x <<= (_23_minus_level);

      int exp = 126; // this corresponds to an exponent of -1

      while( 0 == (x & (1<<23)) )
      {
	 x <<= 1;
	 --exp;
      }

      // clear bit 23
      x &= ~(1<<23);

      // place the biased exponent in the proper place
      exp <<= 23;
      x |= exp;

   }

}


#endif
